from __future__ import annotations
import re
import logging
import sys
import types
import typing
import collections
from dataclasses import MISSING, Field, InitVar

if sys.version_info < (3, 9):
    from typing_extensions import Annotated, get_origin, get_args, ForwardRef
else:
    from typing import Annotated, get_origin, get_args, ForwardRef

from .exceptions import ValidationError

if typing.TYPE_CHECKING:
    from .base import BaseModel

__all__ = (
    "MiniAnnotated",
    "Attrib",
    "get_type",
    "is_collection",
    "is_optional_type",
    "is_type",
    "is_mini_annotated",
    "NoneType",
    "ModelConfigWrapper",
    "is_builtin_type",
    "InitVar",
    "is_initvar_type",
    "is_class_var_type",
    "get_origin",
    "get_args",
)

logger = logging.getLogger(__name__)


# backward compatibility
NoneType = getattr(types, "NoneType", type(None))

COLLECTION_TYPES = frozenset([
    list, tuple, set, frozenset, dict,
    collections.deque, collections.defaultdict,
    collections.Counter, collections.OrderedDict
])



_DATACLASS_CONFIG_FIELD: typing.List[str] = [
    "init",
    "repr",
    "eq",
    "order",
    "unsafe_hash",
    "frozen",
]

_NON_DATACLASS_CONFIG_FIELD: typing.List[str] = [
    "disable_typecheck",
    "disable_all_validation",
]


class ModelConfigWrapper:
    init: bool = True
    repr: bool = True
    eq: bool = True
    order: bool = False
    unsafe_hash: bool = False
    frozen: bool = False
    disable_typecheck: bool = False
    disable_all_validation: bool = False

    def __init__(self, config: typing.Type):
        self.config = config

    def get_config(self, name: str) -> typing.Any:
        if self.config and hasattr(self.config, name):
            return getattr(self.config, name, None)
        return getattr(self.__class__, name, None)

    def get_dataclass_config(self) -> typing.Dict[str, typing.Any]:
        dt = collections.OrderedDict()
        for config_field in _DATACLASS_CONFIG_FIELD:
            dt[config_field] = self.get_config(config_field)
        return dt

    def get_non_dataclass_config(self) -> typing.Dict[str, typing.Any]:
        dt = collections.OrderedDict()
        for config_field in _NON_DATACLASS_CONFIG_FIELD:
            dt[config_field] = self.get_config(config_field)
        return dt


class Attrib:
    __slots__ = (
        "default",
        "default_factory",
        "pre_formatter",
        "required",
        "allow_none",
        "gt",
        "ge",
        "lt",
        "le",
        "min_length",
        "max_length",
        "pattern",
        "_validators",
    )

    def __init__(
        self,
        default: typing.Optional[typing.Any] = MISSING,
        default_factory: typing.Optional[typing.Callable[[], typing.Any]] = MISSING,
        pre_formatter: typing.Callable[[typing.Any], typing.Any] = MISSING,
        required: bool = False,
        allow_none: bool = False,
        gt: typing.Optional[float] = None,
        ge: typing.Optional[float] = None,
        lt: typing.Optional[float] = None,
        le: typing.Optional[float] = None,
        min_length: typing.Optional[int] = None,
        max_length: typing.Optional[int] = None,
        pattern: typing.Optional[typing.Union[str, typing.Pattern]] = None,
        validators: typing.Optional[
            typing.List[typing.Callable[[typing.Any], typing.Any]]
        ] = MISSING,
    ):
        """
        Represents a data attribute with optional validation, default values, and formatting logic.

        Attributes (via __slots__):
            default (Any): A default value for the attribute, if provided.
            default_factory (Callable): A callable that generates a default value.
            pre_formatter (Callable): A function to preprocess/format the value before validation.
            required (bool): Whether the attribute is required.
            allow_none (bool): Whether None is an acceptable value.
            gt (float): Value must be greater than this (exclusive).
            ge (float): Value must be greater than or equal to this (inclusive).
            lt (float): Value must be less than this (exclusive).
            le (float): Value must be less than or equal to this (inclusive).
            min_length (int): Minimum allowed length (for iterable types like strings/lists).
            max_length (int): Maximum allowed length.
            pattern (str or Pattern): Regex pattern the value must match (typically for strings).
            _validators (List[Callable]): Custom validators to run on the value.

        Args:
            default (Any, optional): Static default value to use if none is provided.
            default_factory (Callable, optional): Function that returns a default value.
            pre_formatter (Callable, optional): Function to format/preprocess the value before validation.
            required (bool): Whether this field is required (default: False).
            allow_none (bool): Whether None is allowed as a value (default: False).
            gt, ge, lt, le (float, optional): Numeric comparison constraints.
            min_length, max_length (int, optional): Length constraints for sequences.
            pattern (str or Pattern, optional): Regex pattern constraint.
            validators (List[Callable], optional): Additional callables that validate the input.
        """
        self.default = default
        self.default_factory = default_factory
        self.pre_formatter = pre_formatter
        self.required = required
        self.allow_none = allow_none
        self.gt = gt
        self.ge = ge
        self.lt = lt
        self.le = le
        self.min_length = min_length
        self.max_length = max_length
        self.pattern = pattern

        if validators is not MISSING:
            self._validators = (
                validators if isinstance(validators, (list, tuple)) else [validators]
            )
        else:
            self._validators = []

    def __repr__(self):
        return (
            "Attrib("
            f"default={self.default!r},"
            f"default_factory={self.default_factory!r},"
            ")"
        )

    def has_default(self):
        return self.default is not MISSING or self.default_factory is not MISSING

    def has_pre_formatter(self):
        return self.pre_formatter is not None and callable(self.pre_formatter)

    def _get_default(self) -> typing.Any:
        if self.default is not MISSING:
            return self.default
        elif self.default_factory is not MISSING:
            return self.default_factory()
        return MISSING

    def execute_pre_formatter(self, instance, fd: Field) -> None:
        if self.has_pre_formatter():
            value = getattr(instance, fd.name, None)
            try:
                value = self.pre_formatter(value)
                if self.allow_none and value is None:
                    setattr(instance, fd.name, None)
                else:
                    setattr(instance, fd.name, value)
            except Exception as exc:
                logger.error(
                    "Pre-formatter error for %s : %s", (fd.name, exc), exc_info=exc
                )
                raise RuntimeError(
                    f"Error occurred while executing the pre-formatter for field: {fd.name}"
                ) from exc

    def validate(self, value: typing.Any, field_name: str) -> typing.Optional[bool]:
        value = value or self._get_default()

        if self.allow_none and value is None:
            return True

        if self.required and value is None:
            raise ValidationError(
                f"Field '{field_name}' is required but not provided (value is None).",
                params={"field_name": field_name},
            )

        for name in ("gt", "ge", "lt", "le", "min_length", "max_length", "pattern"):
            validation_factor = getattr(self, name, None)

            # Skip the validation if 'validation_factor' is None, or if both 'value'
            # and 'self.default' are None
            if validation_factor is None or value is None:
                continue

            validator = getattr(self, f"_validate_{name}")
            validator(value)
        return True

    def execute_field_validators(self, instance: "BaseModel", fd: Field) -> None:
        for validator in self._validators:
            try:
                result = validator(instance, getattr(instance, fd.name))
                if result is not None:
                    setattr(instance, fd.name, result)
                elif self.allow_none:
                    setattr(instance, fd.name, None)
            except Exception as e:
                if isinstance(e, ValidationError):
                    raise
                raise ValidationError(str(e)) from e

    def _validate_gt(self, value: typing.Any):
        try:
            if not (value > self.gt):
                raise ValidationError(
                    f"Field value '{value}' is not greater than '{self.gt}'",
                    params={"gt": self.gt},
                )
        except TypeError:
            raise TypeError(
                f"Unable to apply constraint 'gt' to supplied value {value!r}"
            )

    def _validate_ge(self, value: typing.Any):
        try:
            if not (value >= self.ge):
                raise ValidationError(
                    f"Field value '{value}' is not greater than or equal to '{self.ge}'",
                    params={"ge": self.ge},
                )
        except TypeError:
            raise TypeError(
                f"Unable to apply constraint 'ge' to supplied value {value!r}"
            )

    def _validate_lt(self, value: typing.Any):
        try:
            if not (value < self.lt):
                raise ValidationError(
                    f"Field value '{value}' is not less than '{self.lt}'",
                    params={"lt": self.lt},
                )
        except TypeError:
            raise TypeError(
                f"Unable to apply constraint 'lt' to supplied value {value!r}"
            )

    def _validate_le(self, value: typing.Any):
        try:
            if not (value <= self.le):
                raise ValidationError(
                    f"Field value '{value}' is not less than or equal to '{self.le}'",
                    params={"le": self.le},
                )
        except TypeError:
            raise TypeError(
                f"Unable to apply constraint 'le' to supplied value {value!r}"
            )

    def _validate_min_length(self, value: typing.Any):
        try:
            if not (len(value) >= self.min_length):
                raise ValidationError(
                    "too_short",
                    {
                        "field_type": "Value",
                        "min_length": self.min_length,
                        "actual_length": len(value),
                    },
                )
        except TypeError:
            raise TypeError(
                f"Unable to apply constraint 'min_length' to supplied value {value!r}"
            )

    def _validate_max_length(self, value: typing.Any):
        try:
            actual_length = len(value)
            if actual_length > self.max_length:
                raise ValidationError(
                    f"Value is too long. {actual_length} > {self.max_length}",
                    {
                        "field_type": "Value",
                        "max_length": self.max_length,
                        "actual_length": actual_length,
                    },
                )
        except TypeError:
            raise TypeError(
                f"Unable to apply constraint 'max_length' to supplied value {value!r}"
            )

    def _validate_pattern(self, value: typing.Any):
        try:
            if not re.match(self.pattern, value):
                raise ValidationError(
                    f"Field value '{value}' does not match pattern",
                    params={"pattern": self.pattern, "value": value},
                )
        except TypeError:
            raise TypeError(
                f"Unable to apply constraint 'pattern' to supplied value {value!r}"
            )


def is_mini_annotated(typ) -> bool:
    origin = get_origin(typ)
    return (
        origin
        and origin is Annotated
        and hasattr(typ, "__metadata__")
        and Attrib in [inst.__class__ for inst in typ.__metadata__]
    )


def is_type(typ):
    try:
        is_typ = isinstance(typ, type)
    except TypeError:
        is_typ = False
    return is_typ


def is_initvar_type(typ):
    if hasattr(typ, "type"):
        if isinstance(typ, InitVar):
            return typ.__class__.__name__ == "InitVar"
        return hasattr(typ, "__name__") and typ.__name__ == "InitVar"
    return False


def is_class_var_type(typ) -> bool:
    return typ is typing.ClassVar or get_origin(typ) is typing.ClassVar


def get_type(typ):
    if is_type(typ):
        return typ

    if is_optional_type(typ):
        type_args = get_args(typ)
        if type_args:
            return get_type(type_args[0])
        else:
            return

    origin = get_origin(typ)
    if is_type(origin):
        return origin

    type_args = get_args(typ)
    if len(type_args) > 0:
        return get_type(type_args[0])


def is_optional_type(typ):
    if hasattr(typ, "__origin__") and typ.__origin__ is typing.Union:
        return NoneType in typ.__args__
    elif typ is typing.Optional:
        return True
    return False


def is_collection(typ) -> typing.Tuple[bool, typing.Optional[type]]:
    origin = get_origin(typ)
    if origin and origin in COLLECTION_TYPES:
        return True, origin
    return False, None


def get_forward_type(typ):
    """
    Determine if a type annotation is a forward reference and extract the type.

    Args:
        typ: A type annotation that may be a forward reference

    Returns:
        The string name of the forward reference if it is one, otherwise None
    """
    # Check if it's already a string
    if isinstance(typ, str):
        return typ

    # Check if it's a ForwardRef object (Python 3.7+)
    if isinstance(typ, ForwardRef):
        # In Python 3.7-3.10, use __forward_arg__
        # In Python 3.11+, use __arg__
        if sys.version_info >= (3, 11):
            return typ.__arg__
        else:
            return typ.__forward_arg__

    # Check if it's a generic type with forward references (e.g., List['MyClass'])
    origin = get_origin(typ)
    if origin is not None:
        args = get_args(typ)
        # Return the first forward reference found in generic args
        for arg in args:
            forward = get_forward_type(arg)
            if forward:
                return forward

    return None


def is_builtin_type(typ):
    typ = typ if isinstance(typ, type) else type(typ)
    return typ.__module__ in ("builtins", "__builtins__")


class MiniAnnotated:
    __slots__ = ()

    def __init_subclass__(cls, **kwargs):
        raise TypeError(f"Cannot subclass {cls.__module__}.MiniAnnotated")

    def __new__(cls, *args, **kwargs):
        raise TypeError("Type MiniAnnotated cannot be instantiated.")

    @typing._tp_cache
    def __class_getitem__(cls, params):
        if not isinstance(params, tuple):
            params = (params, Attrib())

        if len(params) != 2:
            raise TypeError(
                "MiniAnnotated[...] should be used with exactly two arguments (a type and an Attrib)."
            )

        typ = params[0]

        actual_typ = get_type(typ)
        if actual_typ is None:
            raise ValueError("'{}' is not a type".format(params[0]))

        query = params[1]
        if not isinstance(query, Attrib):
            raise TypeError("Parameter '{}' must be instance of Attrib".format(1))
        return Annotated[typ, query]



class ValidationError(Exception):

    def __init__(self, message, code=None, params=None):
        super().__init__(message)
        self.message = message
        self.code = code
        self.params = params

    def to_dict(self):
        return {
            "error_class": self.__class__.__name__,
            "message": self.message,
            "code": self.code,
            "params": self.params,
        }



"""
===========================================================================================
Copyright (C) 2025 Nafiu Shaibu <nafiushaibu1@gmail.com>.
Purpose: Dataclass with validation
-------------------------------------------------------------------------------------------
This is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 3 of the License, or (at your option)
any later version.

This is distributed in the hopes that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

===========================================================================================
"""

__version__ = "1.1.2"

from .base import BaseModel
from .typing import Attrib, MiniAnnotated


__all__ = ["BaseModel", "Attrib", "MiniAnnotated"]



import typing
import logging
import inspect
from dataclasses import MISSING

from pydantic_mini.typing import is_builtin_type

logger = logging.getLogger(__name__)


T = typing.TypeVar('T')


def get_function_call_args(
    func, params: typing.Union[typing.Dict[str, typing.Any], object]
) -> typing.Dict[str, typing.Any]:
    """
    Extracts the arguments for a function call from the provided parameters.

    Args:
        func: The function for which arguments are to be extracted.
        params: A dictionary of parameters containing
                the necessary arguments for the function.

    Returns:
        A dictionary where the keys are the function argument names
        and the values are the corresponding argument values.
    """
    params_dict = {}
    try:
        sig = inspect.signature(func)
        for param in sig.parameters.values():
            if param.name != "self":
                value = (
                    params.get(param.name, param.default)
                    if isinstance(params, dict)
                    else getattr(params, param.name, param.default)
                )
                if value is not MISSING and value is not inspect.Parameter.empty:
                    params_dict[param.name] = value
                else:
                    params_dict[param.name] = None
    except (ValueError, KeyError) as e:
        logger.warning(f"Parsing {func} for call parameters failed {str(e)}")

    for key in ["args", "kwargs"]:
        if key in params_dict and params_dict[key] is None:
            params_dict.pop(key, None)
    return params_dict


def init_class(
    klass: typing.Type[T], 
    params: typing.Union[typing.Dict[str, typing.Any], object],
    *,
    strict: bool = False,
    allow_extra_attrs: bool = False
) -> T:
    """
    Initialize a class instance with parameters, handling both constructor args and extra attributes.
    
    Args:
        klass: The class to instantiate
        params: Parameters as dictionary or object with attributes
        strict: If True, raise error when extra parameters don't match constructor signature
        allow_extra_attrs: If True, set extra parameters as instance attributes
        
    Returns:
        Initialized instance of the class
        
    Raises:
        TypeError: If class instantiation fails
        ValueError: If strict mode is enabled and extra parameters are found
        AttributeError: If params cannot be converted to dictionary
    """
    if not inspect.isclass(klass):
        raise TypeError(f"Expected a class, got {type(klass)}")
        
    try:
        if hasattr(params, '__dict__'):
            param_dict = params.__dict__.copy()
        elif isinstance(params, dict):
            allow_extra_attrs = False
            param_dict = params.copy()
        else:
            param_dict = vars(params)
    except (TypeError, AttributeError) as e:
        raise AttributeError(f"Cannot extract parameters from {type(params)}: {e}")

    if is_builtin_type(klass):
        return param_dict
    
    try:
        constructor_kwargs = get_function_call_args(klass.__init__, param_dict)
    except ValueError as e:
        raise ValueError(f"Failed to analyze constructor for {klass.__name__}: {e}")
        
    extra_params = {key: value for key, value in param_dict.items() 
                   if key not in constructor_kwargs}
    
    if strict and extra_params:
        extra_keys = ', '.join(extra_params.keys())
        raise ValueError(f"Extra parameters not allowed in strict mode: {extra_keys}")
        
    try:
        instance = klass(**constructor_kwargs)
    except TypeError as e:
        raise TypeError(f"Failed to instantiate {klass.__name__}: {e}")
    
    if allow_extra_attrs and extra_params:
        for key, value in extra_params.items():
            try:
                setattr(instance, key, value)
            except AttributeError as e:
                # Some classes may not allow arbitrary attribute setting
                raise AttributeError(f"Cannot set attribute '{key}' on {klass.__name__}: {e}")
    
    return instance



import unittest
import typing
from unittest.mock import patch
from dataclasses import field, InitVar
from pydantic_mini import BaseModel, MiniAnnotated, Attrib
from pydantic_mini.exceptions import ValidationError


class TestBase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        class MyModel(BaseModel):
            name: str
            age: int

        class DataClassField(BaseModel):
            school = field(default="knust")
            value = field(default_factory=lambda: 1)

        class AnnotatedDataClass(BaseModel):
            email: MiniAnnotated[
                str, Attrib(pattern=r"^[^@]+@[^@]+\.[^@]+$", max_length=13)  # noqa:
            ]
            value: MiniAnnotated[int, Attrib(gt=4, lt=20, default=5)]

        class UsingOptionalDataClass(BaseModel):
            value: typing.Optional[int]
            name: MiniAnnotated[typing.Optional[str], Attrib(max_length=20)]

        class DisabledAllValidationClass(BaseModel):
            email: MiniAnnotated[
                str, Attrib(pattern=r"^[^@]+@[^@]+\.[^@]+$", max_length=13)  # noqa:
            ]
            value: MiniAnnotated[int, Attrib(gt=4, lt=20, default=5)]

            class Config:
                disable_all_validation = True

        class DisabledTypeCheckValidationClass(BaseModel):
            email: MiniAnnotated[
                str, Attrib(pattern=r"^[^@]+@[^@]+\.[^@]+$", max_length=13)  # noqa:
            ]
            value: MiniAnnotated[int, Attrib(gt=4, lt=20, default=5)]

            class Config:
                disable_typecheck = True

        cls.MyModel = MyModel
        cls.DataClassField = DataClassField
        cls.AnnotatedDataClass = AnnotatedDataClass
        cls.UsingOptionalDataClass = UsingOptionalDataClass
        cls.DisabledAllValidationClass = DisabledAllValidationClass
        cls.DisabledTypeCheckValidationClass = DisabledTypeCheckValidationClass

    def test_simple_annotated_model(self):
        instance = self.MyModel(name="test", age=10)
        self.assertEqual(instance.name, "test")
        self.assertEqual(instance.age, 10)

        with self.assertRaises(TypeError):
            self.MyModel(name=12, age="hello")

    def test_dataclass_field(self):
        instance = self.DataClassField()
        self.assertEqual(instance.school, "knust")
        self.assertEqual(instance.value, 1)

        # validate detected type from default
        with self.assertRaises(TypeError):
            self.DataClassField(school=23, value="hello")

    def test_mini_annotated_annotation(self):
        instance = self.AnnotatedDataClass(value=10, email="ex@email.com")
        self.assertEqual(instance.email, "ex@email.com")
        self.assertEqual(instance.value, 10)

        with self.assertRaises(ValidationError):
            self.AnnotatedDataClass(value=10, email="ex")

        with self.assertRaises(ValidationError):
            self.AnnotatedDataClass(value=10, email="looooooooong-email@example.com")

    def test_fields_with_or_without_default_values_cause_error(self):
        class Person(BaseModel):
            name: str = "nafiu"
            school: str

        class Person1(BaseModel):
            name: str
            school: str = "knust"

        p1 = Person(school="knust")
        self.assertEqual(p1.name, "nafiu")
        self.assertEqual(p1.school, "knust")

        p2 = Person1(name="nafiu")
        self.assertEqual(p2.name, "nafiu")
        self.assertEqual(p2.school, "knust")

        # validate positional arguments are required
        with self.assertRaises(TypeError):
            Person(name="nafiu")

        with self.assertRaises(TypeError):
            Person1(school="knust")

    def test_figured_out_optional_field_from_annotation_has_none_value(self):
        p = self.UsingOptionalDataClass()
        self.assertEqual(p.value, None)
        self.assertEqual(p.name, None)

    def test_model_creation_with_dict(self):
        param = {"name": "nafiu", "age": 12}
        instance = self.MyModel.loads(param, _format="dict")
        self.assertEqual(instance.name, "nafiu")
        self.assertEqual(instance.age, 12)

    def test_model_serialization_with_dict(self):
        instance = self.MyModel(name="nafiu", age=12)
        _dict = instance.dump(_format="dict")
        self.assertIsInstance(_dict, dict)
        self.assertEqual(_dict, {"name": "nafiu", "age": 12})

    def test_multiple_model_creation_with_dict(self):
        params = [
            {"name": "nafiu", "age": 12},
            {"name": "shaibu", "age": 13},
            {"name": "nshaibu", "age": 14},
        ]
        instance = self.MyModel.loads(params, _format="dict")
        self.assertIsInstance(instance, list)
        self.assertEqual(len(instance), len(params))

    def test_inner_model_creation_with_dict_and_list(self):
        class School(BaseModel):
            name: str
            location: str

        class Person(BaseModel):
            name: str
            school: typing.List[School]

        instance = Person(name="nafiu", school=[School("knust", location="kumasi")])
        self.assertIsInstance(instance, Person)
        self.assertEqual(instance.name, "nafiu")
        self.assertEqual(instance.school[0].name, "knust")
        self.assertEqual(instance.school[0].location, "kumasi")

        # convert dict to BaseModel test
        instance1 = Person(
            name="shaibu", school=[{"name": "knust", "location": "kumasi"}]
        )
        self.assertIsInstance(instance1, Person)
        self.assertEqual(instance1.name, "shaibu")
        self.assertEqual(instance1.school[0].name, "knust")
        self.assertEqual(instance1.school[0].location, "kumasi")

        # multiple
        instance2 = Person(
            name="nshaibu",
            school=[
                {"name": "knust", "location": "kumasi"},
                School("legon", location="accra"),
            ],
        )
        self.assertIsInstance(instance2, Person)
        self.assertEqual(instance2.name, "nshaibu")
        self.assertEqual(len(instance2.school), 2)

    def test_inner_model_creation_with_normal_class(self):
        class School:
            def __init__(self, name: str, location: str):
                self.name: str = name
                self.location: str = location

        class Student(BaseModel):
            name: str
            school: School

        student = Student(name="nafiu", school=School("knust", location="kumasi"))
        self.assertIsInstance(student, Student)
        self.assertEqual(student.name, "nafiu")
        self.assertIsInstance(student.school, School)
        self.assertEqual(student.school.name, "knust")
        self.assertEqual(student.school.location, "kumasi")

        # multiple entries
        with self.assertRaises(TypeError):
            Student(name="nafiu", school=[School("knust", location="kumasi")])

        class Person(BaseModel):
            name: str
            school: typing.List[School]

        person = Person(name="nafiu", school=[School("knust", location="kumasi")])
        self.assertIsInstance(person, Person)
        self.assertEqual(person.name, "nafiu")
        self.assertIsInstance(person.school, list)
        self.assertEqual(len(person.school), 1)
        self.assertEqual(person.school[0].name, "knust")
        self.assertEqual(person.school[0].location, "kumasi")

        instance1 = Person(
            name="shaibu",
            school=[
                {"name": "knust", "location": "kumasi"},
                School("legon", location="accra"),
            ],
        )
        self.assertIsInstance(instance1, Person)
        self.assertEqual(instance1.name, "shaibu")
        self.assertIsInstance(instance1.school, list)
        self.assertEqual(len(instance1.school), 2)

    def test_validation_for_all_fields(self):
        class Person(BaseModel):
            name: str
            school_name: str

            def validate(self, value, fd):
                if len(value) > 10:
                    raise ValidationError("Value too long")

        instance = Person(name="nafiu", school_name="knust")
        self.assertEqual(instance.school_name, "knust")
        self.assertEqual(instance.name, "nafiu")

        with self.assertRaises(ValidationError):
            Person(name="loooong-user-name", school_name="knust")

        with self.assertRaises(ValidationError):
            Person(name="nafiu", school_name="loooong-school-name")

    def test_custom_field_validator(self):
        def validate_name(instance, value):
            if len(value) > 10:
                raise ValidationError("Value too long")

        def validate_school_name(instance, value):
            if len(value) > 6:
                raise ValidationError("Value too long")

        class Person(BaseModel):
            name: MiniAnnotated[str, Attrib(validators=[validate_name])]
            school_name: MiniAnnotated[str, Attrib(validators=[validate_school_name])]

        person = Person(name="nafiu", school_name="knust")
        self.assertEqual(person.name, "nafiu")
        self.assertEqual(person.school_name, "knust")

        with self.assertRaises(ValidationError):
            Person(
                name="nafiu",
                school_name="kwame nkrumah university of science and technology",
            )

        with self.assertRaises(ValidationError):
            Person(name="very long user name", school_name="knust")

    def test_validators_model_method(self):
        class Person(BaseModel):
            name: str
            school_name: str

            def validate_school_name(self, value, fd):
                if len(value) > 6:
                    raise ValidationError("Value too long")

            def validate_name(self, value, fd):
                if len(value) > 10:
                    raise ValidationError("Value too long")

        person = Person(name="nafiu", school_name="knust")
        self.assertEqual(person.name, "nafiu")
        self.assertEqual(person.school_name, "knust")

        with self.assertRaises(ValidationError):
            Person(
                name="nafiu",
                school_name="kwame nkrumah university of science and technology",
            )

        with self.assertRaises(ValidationError):
            Person(name="verrrry looooooooong naaaaame", school_name="knust")

    def test_model_field_method_validators_can_transform_field_value(self):
        class Person(BaseModel):
            name: str
            school_name: str

            def validate_school_name(self, value, fd):
                return value.upper()

            def validate_name(self, value, fd):
                return value.upper()

        person = Person(name="nafiu", school_name="knust")
        self.assertEqual(person.name, "NAFIU")
        self.assertEqual(person.school_name, "KNUST")

    def test_model_custom_field_validators_can_transform_field_value(self):
        def validate_school_name(instance, value):
            return value.upper()

        def validate_name(instance, value):
            return value.upper()

        class Person(BaseModel):
            name: MiniAnnotated[str, Attrib(validators=[validate_name])]
            school_name: MiniAnnotated[str, Attrib(validators=[validate_school_name])]

        person = Person(name="nafiu", school_name="knust")
        self.assertEqual(person.name, "NAFIU")
        self.assertEqual(person.school_name, "KNUST")

    def test_attribute_validation(self):
        class Person(BaseModel):
            name: MiniAnnotated[str, Attrib(max_length=10, min_length=2)]
            age: MiniAnnotated[int, Attrib(le=14)]
            number_of_dependents: MiniAnnotated[int, Attrib(default=3, lt=4)]

        with self.assertRaises(ValidationError):
            Person(name="n", age=14)

        with self.assertRaises(ValidationError):
            Person(name="very very loong name", age=14)

        with self.assertRaises(ValidationError):
            p0 = Person(name="nshaibu", age=17)
            self.assertEqual(p0.number_of_dependents, 3)

        with self.assertRaises(ValidationError):
            p1 = Person(name="nafiu", age=15, number_of_dependents=2)
            self.assertEqual(p1.number_of_dependents, 2)

    def test_all_field_validator_method_can_transform_field_value(self):
        class Person(BaseModel):
            name: str
            school_name: str

            def validate(self, value, fd):
                return value.upper()

        person = Person(name="nafiu", school_name="knust")
        self.assertEqual(person.name, "NAFIU")
        self.assertEqual(person.school_name, "KNUST")

    def test_union_test_annotation_validator_types_in_union(self):
        class Person(BaseModel):
            name: str
            location: typing.Union[str, int]

        person = Person(name="nafiu", location="kumasi")
        self.assertEqual(person.name, "nafiu")
        self.assertEqual(person.location, "kumasi")

        person = Person(name="nafiu", location=12)
        self.assertEqual(person.name, "nafiu")
        self.assertEqual(person.location, 12)

        with self.assertRaises(TypeError):
            Person(name="nafiu", location=54.554)

    def test_dataclass_composition_association(self):
        class School(BaseModel):
            name: str
            location: str

        class Person(BaseModel):
            name: str
            school: School

        person = Person(name="nafiu", school=School(name="knust", location="kumasi"))
        self.assertEqual(person.name, "nafiu")
        self.assertEqual(person.school.name, "knust")
        self.assertEqual(person.school.location, "kumasi")

        person1 = Person(name="nafiu", school={"name": "knust", "location": "kumasi"})
        self.assertEqual(person1.name, "nafiu")
        self.assertEqual(person1.school.name, "knust")
        self.assertEqual(person1.school.location, "kumasi")

    def test_normal_class_composition_association(self):
        class School:
            def __init__(self, name: str, location: str):
                self.name = name
                self.location = location

        class Person(BaseModel):
            name: str
            school: School

        person = Person(name="nafiu", school=School(name="knust", location="kumasi"))
        self.assertEqual(person.name, "nafiu")
        self.assertEqual(person.school.name, "knust")
        self.assertEqual(person.school.location, "kumasi")

        person1 = Person(name="nafiu", school={"name": "knust", "location": "kumasi"})
        self.assertEqual(person1.name, "nafiu")
        self.assertEqual(person1.school.name, "knust")
        self.assertEqual(person1.school.location, "kumasi")

    def test_init_model_is_call(self):
        class Person(BaseModel):
            name: str
            school: InitVar[str]

            def __model_init__(self, school):
                self.school = school

        person = Person(name="nafiu", school="knust")
        self.assertEqual(person.name, "nafiu")
        self.assertEqual(person.school, "knust")

        with patch.object(Person, "__model_init__", return_value=None) as mock_init:
            Person(name="nafiu", school="knust")
            mock_init.assert_called_once_with("knust")

    def test_positioning_kwargs_before_positional_arguments_does_not_throw_errors(self):
        class Person(BaseModel):
            name: str = "nafiu"
            school: str

        class Person1(BaseModel):
            name: str
            school: str = "knust"

        class Person2(BaseModel):
            name = "nafiu"
            school: int

        person = Person(school="knust")
        self.assertEqual(person.name, "nafiu")
        self.assertEqual(person.school, "knust")

        person1 = Person1(name="nafiu")
        self.assertEqual(person1.name, "nafiu")
        self.assertEqual(person1.school, "knust")

        person2 = Person2(school=33)
        self.assertEqual(person2.name, "nafiu")
        self.assertEqual(person2.school, 33)

        with self.assertRaises(TypeError):
            Person2(name=45, school=32)

    def test_miniannotated_validate_args(self):
        with self.assertRaises(TypeError):

            class Person(BaseModel):
                name: MiniAnnotated[str, 12, "hello"]

        with self.assertRaises(TypeError):

            class Person(BaseModel):
                name: MiniAnnotated[str, 12]

        with self.assertRaises(ValueError):

            class Person(BaseModel):
                name: MiniAnnotated[typing.Optional, Attrib()]

    def test_type_validation_using_any_annotation(self):
        class Person(BaseModel):
            name: typing.Any
            school: typing.Any

        person = Person(name="nafiu", school="knust")
        self.assertEqual(person.name, "nafiu")
        self.assertEqual(person.school, "knust")

    def test_can_validate_collection_fields(self):
        class Person(BaseModel):
            names: typing.List[str]

        person = Person(names=["a", "b", "c"])
        self.assertEqual(person.names, ["a", "b", "c"])

    def test_overridden_init_and_post_init_raises_permissionerror(self):
        with self.assertRaises(PermissionError):

            class Person(BaseModel):
                names: typing.List[str]

                def __init__(self, names):
                    self.names = names

        with self.assertRaises(PermissionError):

            class Person1(BaseModel):
                names: typing.List[str]

                def __post_init__(self):
                    pass

    def test_model_can_be_configured(self):
        class Person(BaseModel):
            name: str
            location: str

            class Config:
                frozen = True

        p = Person(name="nafiu", location="kumasi")
        self.assertEqual(p.name, "nafiu")
        self.assertEqual(p.location, "kumasi")
        self.assertIsInstance(hash(p), int)

        class Person1(BaseModel):
            name: str
            location: str

        p = Person1(name="nafiu", location="kumasi")
        with self.assertRaises(TypeError):
            hash(p)

        class Person2(BaseModel):
            name: str
            location: str

            class Config:
                unsafe_hash = True

        p2 = Person2(name="nafiu", location="kumasi")
        self.assertIsInstance(hash(p2), int)

    def test_disabling_all_validations(self):
        example = self.DisabledAllValidationClass(email="nafiu", value="me")
        self.assertEqual(example.email, "nafiu")
        self.assertEqual(example.value, "me")

    def test_disabling_type_checking(self):
        with self.assertRaises(ValidationError):
            self.DisabledTypeCheckValidationClass(email="nafiu", value="me")

        with self.assertRaises(TypeError):
            self.DisabledTypeCheckValidationClass(email="nafiu@ex.com", value="me")



import typing
import keyword
import inspect
from collections import OrderedDict
from dataclasses import dataclass, fields, Field, field, MISSING, is_dataclass
from .formatters import BaseModelFormatter
from .typing import (
    is_mini_annotated,
    get_type,
    get_origin,
    get_args,
    get_forward_type,
    MiniAnnotated,
    Attrib,
    is_collection,
    is_optional_type,
    is_builtin_type,
    is_initvar_type,
    is_class_var_type,
    ModelConfigWrapper,
)
from .utils import init_class
from .exceptions import ValidationError


__all__ = ("BaseModel",)

PYDANTIC_MINI_EXTRA_MODEL_CONFIG = "__pydantic_mini_extra_config__"


class SchemaMeta(type):

    def __new__(cls, name, bases, attrs, **kwargs):
        parents = [b for b in bases if isinstance(b, SchemaMeta)]
        if not parents:
            return super().__new__(cls, name, bases, attrs)

        cls._prepare_model_fields(attrs)

        new_class = super().__new__(cls, name, bases, attrs, **kwargs)

        model_config_class: typing.Type = getattr(new_class, "Config", None)

        config = ModelConfigWrapper(model_config_class)

        setattr(
            new_class,
            PYDANTIC_MINI_EXTRA_MODEL_CONFIG,
            config.get_non_dataclass_config(),
        )

        return dataclass(new_class, **config.get_dataclass_config())

    @classmethod
    def get_non_annotated_fields(cls, attrs, exclude: typing.Tuple = None):
        if exclude is None:
            exclude = []

        for field_name, value in attrs.items():
            if isinstance(value, (classmethod, staticmethod, property)):
                continue

            # ignore ABC class internal state manager
            if "_abc_impl" == field_name:
                continue

            if (
                not field_name.startswith("__")
                and field_name not in exclude
                and not callable(value)
            ):
                if isinstance(value, Field):
                    typ = cls._figure_out_field_type_by_default_value(
                        field_name, value, attrs
                    )
                else:
                    typ = cls._figure_out_field_type_by_default_value(
                        field_name, value, attrs
                    )
                    value = field(default=value)

                if typ is not None:
                    yield field_name, typ, value

    @classmethod
    def get_fields(
        cls, attrs
    ) -> typing.List[typing.Tuple[typing.Any, typing.Any, typing.Any]]:
        field_dict = {}

        annotation_fields = attrs.get("__annotations__", {})

        for field_name, annotation in annotation_fields.items():
            field_tuple = field_name, annotation
            value = MISSING
            if field_name in attrs:
                value = attrs[field_name]
                value = value if isinstance(value, Field) else field(default=value)

            field_tuple = (*field_tuple, value)

            field_dict[field_name] = field_tuple

        # get fields without annotation
        for field_name, annotation, value in cls.get_non_annotated_fields(
            attrs, exclude=tuple(field_dict.keys())
        ):
            field_dict[field_name] = field_name, annotation, value

        return list(field_dict.values())

    @classmethod
    def _figure_out_field_type_by_default_value(
        cls, field_name: str, value: Field, attrs: typing.Dict[str, typing.Any]
    ) -> typing.Any:
        if isinstance(value, Field):
            if value.default is not MISSING:
                return type(value.default)
            elif value.default_factory is not MISSING:
                return type(value.default_factory())
        elif hasattr(value, "__class__"):
            return value.__class__
        else:
            if field_name in attrs:
                return type(value)
        return typing.Any

    @classmethod
    def _prepare_model_fields(cls, attrs: typing.Dict[str, typing.Any]) -> None:
        ann_with_defaults = OrderedDict()
        ann_without_defaults = OrderedDict()

        for field_name, annotation, value in cls.get_fields(attrs):
            if not isinstance(field_name, str) or not field_name.isidentifier():
                raise TypeError(
                    f"Field names must be valid identifiers: {field_name!r}"
                )
            if keyword.iskeyword(field_name):
                raise TypeError(f"Field names must not be keywords: {field_name!r}")

            if annotation is None:
                if value not in (MISSING, None):
                    annotation = cls._figure_out_field_type_by_default_value(
                        field_name, value, attrs
                    )

                if annotation is None:
                    raise TypeError(
                        f"Field '{field_name}' does not have type annotation. "
                        f"Figuring out field type from default value failed"
                    )

            if (
                is_initvar_type(annotation)
                or is_class_var_type(annotation)
                or annotation is typing.Any
            ):
                # let's ignore init-var and class-var, dataclass will take care of them
                # typing.Any does not require any type Validation
                ann_with_defaults[field_name] = annotation
                if field_name in attrs:
                    value = attrs[field_name]
                    attrs[field_name] = (
                        value.default if isinstance(value, Field) else value
                    )
                continue

            if not is_mini_annotated(annotation):
                if get_type(annotation) is None:
                    # Let's confirm that the annotation isn't a forward type
                    # Forward Types are only for static analyzers and thus cannot
                    # be used at runtime for validation so essentially we will annotate
                    # the filed with typing.Any
                    forward_annotation = get_forward_type(annotation)
                    if forward_annotation is None:
                        raise TypeError(
                            f"Field '{field_name}' must be annotated with a real type. {annotation} is not a type"
                        )
                    else:
                        annotation = typing.Any

                annotation = MiniAnnotated[
                    annotation,
                    Attrib(
                        default=value.default if isinstance(value, Field) else value,
                        default_factory=(
                            value.default_factory if isinstance(value, Field) else value
                        ),
                    ),
                ]

            annotation_type = annotation.__args__[0]
            attrib = annotation.__metadata__[0]

            if is_optional_type(annotation_type):
                # all optional annotations without default value will have
                # None as default
                if not attrib.has_default():
                    attrib.default = None
                    attrs[field_name] = field(default=None)

            if value is MISSING:
                if attrib.has_default():
                    if attrib.default is not MISSING:
                        attrs[field_name] = field(default=attrib.default)
                    else:
                        attrs[field_name] = field(
                            default_factory=attrib.default_factory
                        )

            if attrib.has_default():
                ann_with_defaults[field_name] = annotation
            else:
                ann_without_defaults[field_name] = annotation

        ann_without_defaults.update(ann_with_defaults)

        if ann_without_defaults:
            attrs["__annotations__"] = ann_without_defaults


class PreventOverridingMixin:

    _protect = ["__init__", "__post_init__"]

    def __init_subclass__(cls, **kwargs):
        if cls.__name__ != "BaseModel":
            for attr_name in cls._protect:
                if attr_name in cls.__dict__:
                    raise PermissionError(
                        f"Model '{cls.__name__}' cannot override {attr_name!r}. "
                        f"Consider using __model_init__ for all your custom initialization"
                    )
        super().__init_subclass__(**kwargs)


class BaseModel(PreventOverridingMixin, metaclass=SchemaMeta):

    def __model_init__(self, *args, **kwargs) -> None:
        pass

    def __post_init__(self, *args, **kwargs) -> None:
        config = getattr(self, PYDANTIC_MINI_EXTRA_MODEL_CONFIG, {})
        disable_typecheck = config.get("disable_typecheck", False)
        disable_all_validation = config.get("disable_all_validation", False)

        for fd in fields(self):
            field_type = fd.type
            query: Attrib = (
                hasattr(field_type, "__metadata__")
                and field_type.__metadata__[0]
                or None
            )
            if query:
                # execute the pre-formatters for all the fields
                query.execute_pre_formatter(self, fd)

            if not disable_all_validation:
                # no type validation for Any field type and type checking is not disabled
                if field_type is not typing.Any and not disable_typecheck:
                    self._inner_schema_value_preprocessor(fd)
                    self._field_type_validator(fd)
                else:
                    # run other field validators when type checking is disabled
                    if query:
                        value = getattr(self, fd.name, None)
                        query.execute_field_validators(self, fd)
                        query.validate(value, fd.name)
                try:
                    result = self.validate(getattr(self, fd.name), fd)
                    if result is not None:
                        setattr(self, fd.name, result)
                except NotImplementedError:
                    pass

                method = getattr(self, f"validate_{fd.name}", None)
                if method and callable(method):
                    result = method(getattr(self, fd.name), fd)
                    if result is not None:
                        setattr(self, fd.name, result)

        self.__model_init__(*args, **kwargs)

    def _inner_schema_value_preprocessor(self, fd: Field):
        value = getattr(self, fd.name)
        field_type = fd.type

        actual_annotated_type = field_type.__args__[0]
        type_args = (
            hasattr(actual_annotated_type, "__args__")
            and actual_annotated_type.__args__
            or None
        )

        status, actual_type = is_collection(actual_annotated_type)
        if status:
            if type_args and isinstance(value, (dict, list)):
                value = value if isinstance(value, list) else [value]
                inner_type: type = type_args[0]
                if is_builtin_type(inner_type):
                    setattr(
                        self, fd.name, actual_type([inner_type(val) for val in value])
                    )
                elif (
                    isinstance(inner_type, BaseModel)
                    or is_dataclass(inner_type)
                    or inspect.isclass(inner_type)
                ):
                    setattr(
                        self,
                        fd.name,
                        actual_type(
                            [
                                (
                                    init_class(inner_type, val)
                                    if isinstance(val, dict)
                                    else val
                                )
                                for val in value
                            ]
                        ),
                    )
        elif type_args is None and actual_annotated_type:
            actual_annotated_type = get_type(actual_annotated_type)
            if (
                isinstance(actual_annotated_type, BaseModel)
                or is_dataclass(actual_annotated_type)
                or inspect.isclass(actual_annotated_type)
            ):
                if isinstance(value, dict):
                    setattr(self, fd.name, init_class(actual_annotated_type, value))

    def _field_type_validator(self, fd: Field):
        value = getattr(self, fd.name, None)
        field_type = fd.type

        if not is_mini_annotated(field_type):
            raise ValidationError(
                "Field '{}' should be annotated with 'MiniAnnotated'.".format(fd.name),
                params={"field": fd.name, "annotation": field_type},
            )

        query = field_type.__metadata__[0]

        if not query.has_default() and value is None:
            raise ValidationError(
                "Field '{}' should not be empty.".format(fd.name),
                params={"field": fd.name, "annotation": field_type},
            )

        query.execute_field_validators(self, fd)

        expected_annotated_type = (
            hasattr(field_type, "__args__") and field_type.__args__[0] or None
        )
        actual_expected_type = (
            expected_annotated_type
            and self.type_can_be_validated(expected_annotated_type)
            or None
        )

        if expected_annotated_type and typing.Any not in actual_expected_type:
            is_type_collection, _ = is_collection(expected_annotated_type)
            if is_type_collection:
                actual_type = expected_annotated_type.__args__[0]
                if actual_type and actual_type is not typing.Any:
                    if any([not isinstance(val, actual_type) for val in value]):
                        raise TypeError(
                            "Expected a collection of values of type '{}'. Values: {} ".format(
                                actual_type, value
                            )
                        )
            elif not isinstance(value, actual_expected_type):
                raise TypeError(
                    f"Field '{fd.name}' should be of type {actual_expected_type}, "
                    f"but got {type(value).__name__}."
                )

        query.validate(value, fd.name)

    @staticmethod
    def type_can_be_validated(typ) -> typing.Optional[typing.Tuple]:
        origin = get_origin(typ)
        if origin is typing.Union:
            type_args = get_args(typ)
            if type_args:
                return tuple([get_type(_type) for _type in type_args])
        else:
            return (get_type(typ),)

    @staticmethod
    def get_formatter_by_name(name: str) -> BaseModelFormatter:
        return BaseModelFormatter.get_formatter(format_name=name)

    def validate(self, value: typing.Any, data_field: Field):
        """Implement this method to validate all fields"""
        raise NotImplementedError

    @classmethod
    def loads(
        cls, data: typing.Any, _format: str
    ) -> typing.Union[typing.List["BaseModel"], "BaseModel"]:
        return cls.get_formatter_by_name(_format).encode(cls, data)

    def dump(self, _format: str) -> typing.Any:
        return self.get_formatter_by_name(_format).decode(instance=self)
